;; Generated package description from buttercup.el  -*- no-byte-compile: t; lexical-binding:t -*-
(define-package "buttercup" "1.39.0.20260409.235427" "Behavior-Driven Emacs Lisp Testing" '((emacs "24.4")) :commit "7baf9e5a60f201542309238d77aa6bb35bd77ec0" :authors '(("Jorgen Schaefer" . "contact@jorgenschaefer.de")) :maintainer '("Ola Nilsson" . "ola.nilsson@gmail.com") :url "https://github.com/jorgenschaefer/emacs-buttercup")
